import gzip
import io

import pandas as pd

from ncbitax2lin import utils


def strip(str_: str) -> str:
    """
    :param str_: a string
    """
    return str_.strip()


@utils.timeit
def load_nodes(nodes_file: str) -> pd.DataFrame:
    """
    load nodes.dmp and convert it into a pandas.DataFrame
    """
    df = pd.read_csv(
        nodes_file,
        sep="|",
        header=None,
        index_col=False,
        names=[
            "tax_id",
            "parent_tax_id",
            "rank",
            "embl_code",
            "division_id",
            "inherited_div_flag",
            "genetic_code_id",
            "inherited_GC__flag",
            "mitochondrial_genetic_code_id",
            "inherited_MGC_flag",
            "GenBank_hidden_flag",
            "hidden_subtree_root_flag",
            "comments",
        ],
    )

    # To get rid of flanking tab characters
    df["rank"] = df["rank"].apply(strip)
    df["embl_code"] = df["embl_code"].apply(strip)
    df["comments"] = df["comments"].apply(strip)
    return df


@utils.timeit
def load_names(names_file: str) -> pd.DataFrame:
    """
    load names.dmp and convert it into a pandas.DataFrame
    """
    df = pd.read_csv(
        names_file,
        sep="|",
        header=None,
        index_col=False,
        names=["tax_id", "name_txt", "unique_name", "name_class"],
    )
    df["name_txt"] = df["name_txt"].apply(strip)
    df["unique_name"] = df["unique_name"].apply(strip)
    df["name_class"] = df["name_class"].apply(strip)

    sci_df = df[df["name_class"] == "scientific name"]
    sci_df.reset_index(drop=True, inplace=True)
    return sci_df


def read_names_and_nodes(names_file: str, nodes_file: str) -> pd.DataFrame:
    # data downloaded from ftp://ftp.ncbi.nih.gov/pub/taxonomy/
    # args = parse_args()
    nodes_df = load_nodes(nodes_file)
    names_df = load_names(names_file)

    return (
        nodes_df.merge(names_df, on="tax_id")[
            ["tax_id", "parent_tax_id", "rank", "name_txt"]
        ]
        .rename(columns={"name_txt": "rank_name"})
        .reset_index(drop=True)
    )


def write_lineages_to_disk(df_lineages: pd.DataFrame, output_path: str) -> None:
    """Gzip lineages and write them to disk"""
    cols = [
        "tax_id",
        "superkingdom",
        "phylum",
        "class",
        "order",
        "family",
        "genus",
        "species",
    ]
    other_cols = sorted([col for col in df_lineages.columns if col not in cols])
    output_cols = cols + other_cols

    df_lineages.to_csv(
        output_path, index=False, compression="gzip", columns=output_cols
    )



import argparse
import gzip
import logging
import multiprocessing
import os
import re
from typing import Container, Dict, Iterable, List, NewType, Optional, Tuple, Union

import fire
import pandas as pd
from typing_extensions import TypedDict

from ncbitax2lin import data_io, utils

logging.basicConfig(level=logging.DEBUG, format="%(asctime)s|%(levelname)s|%(message)s")


class TaxUnit(TypedDict):
    """
    Represents a basic unit in taxonomy e.g. (phylum, Proteobacteria), where
    phylum is the rank, and Proteobacteria is the rank name
    """

    tax_id: int
    # tax_id of parent tax unit for this tax unit
    parent_tax_id: int
    rank: str
    rank_name: str


# the strings are tax_id, rank, rank_name
Lineage = NewType("Lineage", List[Tuple[int, str, str]])

# set TAXONOMY_DICT as global variable so it can work with multiprocess.Pool
# more easily
TAXONOMY_DICT: Dict[int, TaxUnit] = {}

# tax_id of first line in names.dmp: no rank
ROOT_TAX_ID = 1


def calc_rank_key(rank: str, existing_ranks: Container[str]) -> str:
    """Calcluates a key for the lineage representation in a dictionary.

    Defaults to the rank itself, e.g. no rank, superkingdom, phylum, etc. but
    when a rank appears multiple times (common for "no rank" rank) in a single
    linearge it will be numbered, e.g. no rank1, no rank2, and so on.

    Args:
        rank: e.g. no rank, superkingdom, phylum, etc.
        existing_ranks: rank keys already existing
    """
    # e.g. there could be multiple 'no rank'
    if rank not in existing_ranks:
        return rank
    else:
        count = 1
        numbered_rank = f"{rank}{count}"
        while numbered_rank in existing_ranks:
            count += 1
            numbered_rank = f"{rank}{count}"
        return numbered_rank


def calc_taxonomy_dict(df_tax: pd.DataFrame) -> Dict[int, TaxUnit]:
    """Converts dataframe of df_tax into a dictionary with tax_id as the keys"""
    return dict(zip(df_tax.tax_id.values, df_tax.to_dict("records")))


def find_lineage(tax_id: int) -> Lineage:
    if tax_id % 50000 == 0:
        logging.debug("working on tax_id: {0}".format(tax_id))

    lineage = []
    while True:
        record = TAXONOMY_DICT[tax_id]
        lineage.append((record["tax_id"], record["rank"], record["rank_name"]))
        tax_id = record["parent_tax_id"]

        # every tax can be traced back to tax_id == 1, the root
        if tax_id == ROOT_TAX_ID:
            break

    # reverse results in lineage of Kingdom => species, this is helpful for
    # to_dict when there are multiple "no rank"s
    lineage.reverse()
    return Lineage(lineage)


def find_all_lineages(tax_ids: Iterable) -> List[Lineage]:
    """find the lineages for all tax ids"""
    ncpus = multiprocessing.cpu_count()
    logging.info(
        "found {0} cpus, and will use all of them to find lineages "
        "for all tax ids".format(ncpus)
    )

    with multiprocessing.Pool(ncpus) as pool:
        return pool.map(find_lineage, tax_ids)


def convert_lineage_to_dict(lineage: Lineage) -> Dict[str, Union[int, str]]:
    """Converts the lineage in a list-of-tuples represetantion to a dictionary representation

    [
        (tax_id1, rank1, name_txt1),
        (tax_id2, rank2, name_txt2),
        ...
    ]

    becomes

    {
        rank1: name_txt1,
        rank2: name_txt2,
        tax_id, tax_id2,   # using the last rank as the tax_id of this lineage
    }

    A concrete example:

        [
            (131567, 'no rank', 'cellular organisms'),
            (2, 'superkingdom', 'Bacteria')
        ]

    becomes

        {
            'no rank': 'cellular organisms',
            'tax_id': 2,
            'superkingdom': 'Bacteria'
        }

    """
    output: Dict[str, Union[int, str]] = {}
    len_lineage = len(lineage)
    for k, (tax_id, rank, rank_name) in enumerate(lineage):
        # use the last rank of the lineage as the tax_id of the lineage
        if k == len_lineage - 1:
            output["tax_id"] = tax_id

        rank_key = calc_rank_key(rank, output.keys())
        output[rank_key] = rank_name
    return output


def prepare_lineages_for_output(lineages: List[Lineage]) -> pd.DataFrame:
    """prepares lineages into a dataframe for writing to disk"""
    logging.info("Preparings all lineages into a dataframe to be written to disk ...")

    df_out = pd.DataFrame([convert_lineage_to_dict(lineage) for lineage in lineages])

    return df_out.sort_values("tax_id")


def taxonomy_to_lineages(
    nodes_file: str, names_file: str, output: Optional[str] = None
) -> None:
    """Converts NCBI taxomony dump into lineages.

    NCBI taxonomy dump can be downloaded from
    ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump.tar.gz

    Args:
        nodes_file: path/to/taxdump/nodes.dmp from NCBI taxonomy
        names_file: path/to/taxdump/names.dmp from NCBI taxonomy
        output_prefix: output lineages will be written to output_prefix.csv.gz
    """
    df_data = data_io.read_names_and_nodes(names_file, nodes_file)
    logging.info(f"# of tax ids: {df_data.shape[0]:,d}")
    logging.info(f"df.info:\n{utils.collect_df_info(df_data)}")

    logging.info("Generating TAXONOMY_DICT ...")
    global TAXONOMY_DICT
    TAXONOMY_DICT = calc_taxonomy_dict(df_data)

    lineages = find_all_lineages(df_data.tax_id)

    df_lineages = prepare_lineages_for_output(lineages)

    if output is None:
        output = f"ncbi_lineages_{pd.Timestamp.utcnow().date()}.csv.gz"
    utils.maybe_backup_file(output)
    logging.info(f"Writing lineages to {output} ...")
    data_io.write_lineages_to_disk(df_lineages, output)


def main() -> None:
    fire.Fire(taxonomy_to_lineages)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ncbitax2lin']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.3.1,<0.4.0', 'pandas>=1.0.3,<2.0.0']

entry_points = \
{'console_scripts': ['poetry = ncbitax2lin.ncbitax2lin:main']}

setup_kwargs = {
    'name': 'ncbitax2lin',
    'version': '2.0.0a1',
    'description': 'A tool that converts NCBI taxonomy dump into lineages',
    'long_description': '# NCBItax2lin\n\nConvert NCBI taxonomy dump from ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump.tar.gz\ninto lineages. An example for [human (tax_id=9606)](https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=9606) is like\n\n| tax_id | superkingdom | phylum   | class    | order    | family    | genus | species      | family1 | forma | genus1 | infraclass | infraorder  | kingdom | no rank            | no rank1     | no rank10            | no rank11 | no rank12 | no rank13 | no rank14 | no rank15     | no rank16 | no rank17 | no rank18 | no rank19 | no rank2  | no rank20 | no rank21 | no rank22 | no rank3  | no rank4      | no rank5   | no rank6      | no rank7   | no rank8     | no rank9      | parvorder  | species group | species subgroup | species1 | subclass | subfamily | subgenus | subkingdom | suborder    | subphylum | subspecies | subtribe | superclass | superfamily | superorder       | superorder1 | superphylum | tribe | varietas |\n|--------|--------------|----------|----------|----------|-----------|-------|--------------|---------|-------|--------|------------|-------------|---------|--------------------|--------------|----------------------|-----------|-----------|-----------|-----------|---------------|-----------|-----------|-----------|-----------|-----------|-----------|-----------|-----------|-----------|---------------|------------|---------------|------------|--------------|---------------|------------|---------------|------------------|----------|----------|-----------|----------|------------|-------------|-----------|------------|----------|------------|-------------|------------------|-------------|-------------|-------|----------|\n| 9606   | Eukaryota    | Chordata | Mammalia | Primates | Hominidae | Homo  | Homo sapiens |         |       |        |            | Simiiformes | Metazoa | cellular organisms | Opisthokonta | Dipnotetrapodomorpha | Tetrapoda | Amniota   | Theria    | Eutheria  | Boreoeutheria |           |           |           |           | Eumetazoa |           |           |           | Bilateria | Deuterostomia | Vertebrata | Gnathostomata | Teleostomi | Euteleostomi | Sarcopterygii | Catarrhini |               |                  |          |          | Homininae |          |            | Haplorrhini | Craniata  |            |          |            | Hominoidea  | Euarchontoglires |             |             |       |          |\n\n## Regenerate the lineages yourself\n\nRegeneration is straightforward, but it may incur quite a bit of memory (~20\nGB). I generated `lineages.csv.gz` on a machine with 32 GB memory. Pull request\non refactoring to a lower memory usage is welcome. It\'s mainly about\n[this line](https://github.com/zyxue/ncbitax2lin/blob/master/ncbitax2lin.py#L184),\nwhere the `pool.map` takes places.\n\nIf you really need an updated version but without the hardware resources, you\ncould also notify me on github, and I will update it for you.\n\n### Install\n\n```\ngit clone git@github.com:zyxue/ncbitax2lin.git\ncd ncbitax2lin/\n```\n\n#### Set up a virtual environment\n\nCurrently, it only works with `python2.7`, and needs\n[pandas](http://pandas.pydata.org/), so make sure you are in a proper virtual\nenvironment. If you have already these had one available, just use that\none.\n\nOtherwise, you can create a new one with\n[miniconda](https://conda.io/miniconda.html)/[anaconda](https://www.continuum.io/downloads)\n(recommended),\n\n```\nconda create -y -p venv/ --file env-conda.txt\n# or effectively the same\n# conda create -y -p venv python=2 pandas\nsource activate venv/\n```\n\nor with [virtualenv + pip](http://docs.python-guide.org/en/latest/dev/virtualenvs/)\n\n```\nvirtualenv venv/\nsource venv/bin/activate\npip install -r env-pip.txt\n```\n\n### Regenerate\nThen run the following, this will download the latest taxdump from NCBI, and run\nthe scripts to regenerate all latest lineages from it\n\n```\nmake\n```\n\n## FAQ\n\n**Q**: I have a large number of sequences with their corresponding accession\nnumbers from NCBI, how to get their lineages?\n\n**A**: First, you need to map accession numbers (GI is deprecated) to tax IDs\nbased on `nucl_*accession2taxid.gz` files from\nftp://ftp.ncbi.nih.gov/pub/taxonomy/accession2taxid/. Secondly, you can trace a\nsequence\'s whole lineage based on its tax ID. The tax-id-to-lineage mapping is\nwhat NCBItax2lin generates for you, and it is available on <a\nhref="https://gitlab.com/zyxue/ncbitax2lin-lineages/tree/master"\ntarget="_blank">ncbitax2lin-lineages</a>.\n\nIf you have any question about this project, please feel free to create a new\n[issue](https://github.com/zyxue/ncbitax2lin/issues/new).\n\n## Note on `taxdump.tar.gz.md5`\n\nIt appears that NCBI periodically regenerates `taxdump.tar.gz` and\n`taxdump.tar.gz.md5` even when its content is still the same. I am not sure how\ntheir regeneration works, but `taxdump.tar.gz.md5` will differ simply because\nof a different timestamp.\n\n## Used in\n\n* Mahmoudabadi, G., & Phillips, R. (2018). A comprehensive and quantitative exploration of thousands of viral genomes. ELife, 7. https://doi.org/10.7554/eLife.31955\n',
    'author': 'Zhuyi Xue',
    'author_email': 'zhuyi@alum.utoronto.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zyxue/ncbitax2lin',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import datetime
import functools
import io
import logging
import os
import time
from typing import Any, Callable, Optional

import pandas as pd

_LOGGER = logging.getLogger(__name__)


def timeit(func: Callable[..., Any]) -> Callable[..., Any]:
    """Times a function, usually used as decorator"""

    @functools.wraps(func)
    def timed_func(*args: Any, **kwargs: Any) -> Any:
        """Returns the timed function"""
        start_time = time.time()
        result = func(*args, **kwargs)
        elapsed_time = datetime.timedelta(seconds=(time.time() - start_time))
        _LOGGER.info("time spent on %s: %s", func.__name__, elapsed_time)
        return result

    return timed_func


def maybe_backup_file(filepath: str) -> None:
    """
    Back up a file, old_file will be renamed to #old_file.n#, where n is a
    number incremented each time a backup takes place
    """
    backup = None
    if os.path.exists(filepath):
        dirname = os.path.dirname(filepath)
        basename = os.path.basename(filepath)
        count = 1
        backup = os.path.join(dirname, f"#{basename}.{count}#")
        while os.path.exists(backup):
            count += 1
            backup = os.path.join(dirname, f"#{basename}.{count}#")
        logging.info(f"Backing up {filepath} to {backup}")
        os.rename(filepath, backup)


def collect_df_info(df_data: pd.DataFrame) -> str:
    """Collects information of a dataframe"""
    buf = io.StringIO()
    df_data.info(buf=buf, verbose=True, memory_usage="deep")
    return buf.getvalue()
